from __future__ import print_function
import click
import gnupg
import pkg_resources
import re
import six
import sys
import tempfile
import ballsy.config
import ballsy.options
import ballsy.login

CONFIG = ballsy.config.Config()


def check_key(gpg, keyid):
    if gpg.export_keys(keyid, True):
        return
    raise RuntimeError(
        "Invalid keyid: {0} -- not found in keyring".format(keyid))


def repo_split(repo):
    p = re.compile('([^/]+)/(.+)')
    m = p.match(repo)
    if m:
        ruser = m.group(1)
        rrepo = m.group(2)
    else:
        print("Invalid repository name: {0}".format(repo), file=sys.stderr)
        sys.exit(1)
    return ruser, rrepo


def build_formats(only_targz, only_zip):
    formats = {}
    if only_targz:
        formats = {'tarball': 'tar.gz'}
    elif only_zip:
        formats = {'zipball': 'zip'}
    else:
        formats = {'tarball': 'tar.gz', 'zipball': 'zip'}
    return formats


@click.group()
@click.option('--verbose', '-v', is_flag=True, help="Be verbose.")
@click.version_option(version=pkg_resources.require("ballsy")[0].version)
@click.pass_context
def main(ctx, verbose):
    """GitHub release signing tool"""
    ctx.verbose = verbose
    pass


@main.command()
@click.option('--key-id', '-k',
              help='Key ID to use for signing. (default: {0})'.format(
                  CONFIG.git_config('user.signingkey')),
              required=True, default=CONFIG.git_config('user.signingkey'),
              metavar='<keyid>')
@click.option('--only-zip', '-z', is_flag=True,
              cls=ballsy.options.MutuallyExclusiveOption,
              help='Only sign the ZIP archive for release.',
              mutually_exclusive=["only-targz"])
@click.option('--only-targz', '-g', is_flag=True,
              cls=ballsy.options.MutuallyExclusiveOption,
              help='Only sign the tarball for release.',
              mutually_exclusive=["only-zip"])
@click.option('--include-tags', '-t', is_flag=True,
              help='Create new release for tag if not present.')
@click.option('--no-draft', '-d', is_flag=True,
              help='Do not set draft flag for new releases.')
@click.option('--prerelease', '-p', is_flag=True,
              help='Set prerelease flag for new releases.')
@click.option('--force', '-f', is_flag=True,
              help='Force re-signing even when release has signature.')
@click.option('--repo', '-r', metavar='<user/repository>',
              cls=ballsy.options.MutuallyExclusiveOption,
              mutually_exclusive=["remote"],
              help='Repository to sign releases for.')
@click.option('--remote', '-m', cls=ballsy.options.MutuallyExclusiveOption,
              metavar='<remote>', mutually_exclusive=["repo"],
              default='origin',
              help='Remote specifying repository to sign releases for. ' +
                   ' (default: origin)')
@click.argument('tag', required=True, nargs=-1)
@click.pass_context
def sign(ctx, key_id, only_zip, only_targz, include_tags, no_draft, prerelease,
         force, repo, remote, tag):
    """Sign release(s) and tags."""

    if remote and not repo and CONFIG.remotes[remote]:
        ruser = CONFIG.remotes[remote][0]
        rrepo = CONFIG.remotes[remote][1]
    else:
        ruser, rrepo = repo_split(repo)

    try:
        gh = ballsy.login.login_with_token(CONFIG.token())
        repo = gh.repository(ruser, rrepo)
        gpg = gnupg.GPG()
        check_key(gpg, key_id)
        for t in tag:
            r = repo.release_from_tag(t)
            if not r:
                gtag = repo.ref('tags/' + t)
                if not (gtag and 'refs/tags/{0}'.format(t) in str(gtag)):
                    print("tag {0} not found, skipping".format(t),
                          file=sys.stderr)
                    continue
                assert(gtag)
                if include_tags:
                    r = repo.create_release(t, gtag.as_dict()['object']['sha'],
                                            draft=(not no_draft),
                                            prerelease=prerelease)
                else:
                    print(('tag {0} found but --include-tags not given, ' +
                          'not creating release').format(t), file=sys.stderr)
                    continue
            assert(r)
            for f, ext in six.iteritems(build_formats(only_targz, only_zip)):
                sigfilename = "{0}.{1}.asc".format(t, ext)
                tb = tempfile.TemporaryFile()
                r.archive(f, path=tb)
                for a in r.assets():
                    if a.name == sigfilename:
                        a.delete()
                signed_data = gpg.sign_file(tb, keyid=key_id, detach=True)
                print(sigfilename)
                with tempfile.TemporaryFile() as outfile:
                    outfile.write(signed_data.data)
                    outfile.seek(0)
                    r.upload_asset('text/ascii', sigfilename, outfile)
    except Exception as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)


@main.command()
@click.pass_context
def login(ctx):
    """Log into GitHub and store user credentials."""
    try:
        g, user, pwd = ballsy.login.login_with_password()
        if CONFIG.has_token():
            ballsy.login.logout(g, CONFIG)
        auth = g.authorize(user, pwd, ['user', 'repo'], "Ballsy")
        CONFIG.set_token(auth.id, auth.token)
    except RuntimeError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)


@main.command()
@click.pass_context
def logout(ctx):
    """Log out of GitHub and clear user credentials."""
    try:
        g, user, pwd = ballsy.login.login_with_password()
        ballsy.login.logout(g, CONFIG)
    except RuntimeError as e:
        print(str(e), file=sys.stderr)
        sys.exit(1)



import getpass
import github3


def login_with_password():
    try:
        prompt = raw_input
    except NameError:
        prompt = input

    user = ''
    while not user:
        user = prompt('GitHub username: ')

    pwd = ''
    while not pwd:
        pwd = getpass.getpass('GitHub password for {0}: '.format(user))

    def twofa_auth():
        code = ''
        while not code:
            code = prompt('Enter 2FA code: ')
        return code

    g = github3.login(user, pwd, two_factor_callback=twofa_auth)
    return g, user, pwd


def login_with_token(token):
    g = github3.login(token=token)
    return g


def logout(github, config):
    if config.has_token():
        a = github.authorization(config.id())
        a.delete()
    config.cfg['login'] = None



import configparser
import os
import re
import six
import stat
import subprocess

GITCONFIGLINE = re.compile('^([^=]+)=(.+)$')
GITREMOTEKEY = re.compile('^remote.([^.]+).url')


class Config(object):

    def _extract_github_repo(self, url):
        m = re.match("^git@github.com:([a-zA-Z0-9_-]+)/([a-zA-Z0-9_-]+).git$",
                     url) or \
            re.match("^https://github.com/([a-zA-Z0-9_-]+)/([a-zA-Z0-9_-]+)" +
                     ".git$", url)
        return (m.group(1), m.group(2)) if m else None

    def _get_remotes(self):
        self.remotes = {}
        for k, v in six.iteritems(self.gitcfg):
            m = GITREMOTEKEY.match(k)
            if m:
                self.remotes[m.group(1)] = self._extract_github_repo(v)

    def __init__(self):
        self.fn = os.path.join(os.path.expanduser('~'), '.ballsyrc')
        self.cfg = configparser.ConfigParser()
        self.gitcfg = {}
        if os.path.isfile(self.fn):
            self.cfg.read(self.fn)
        try:
            process = subprocess.Popen(["git", "config", "-l"],
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            out, err = process.communicate()
            for l in out.splitlines():
                m = GITCONFIGLINE.match(l.decode("utf-8"))
                if m:
                    self.gitcfg[m.group(1)] = m.group(2)
        except:
            pass
        self._get_remotes()

    def git_config(self, val):
        if val in self.gitcfg:
            return self.gitcfg[val]
        else:
            return None

    def has_token(self):
        if 'login' in self.cfg:
            if 'id' not in self.cfg['login'] or len(
                    self.cfg['login']['id']) == 0:
                return False
            if 'token' not in self.cfg['login'] or len(
                    self.cfg['login']['token']) == 0:
                return False
            return True
        else:
            return False

    def set_token(self, id, token):
        self.cfg['login'] = {'id': id, 'token': token}
        with open(self.fn, 'w+') as configfile:
            self.cfg.write(configfile)
        os.chmod(self.fn, stat.S_IRWXU)

    def token(self):
        if 'login' not in self.cfg:
            raise RuntimeError('No token found in ' + self.fn +
                               ", please run 'ballsy login'")
        else:
            return self.cfg['login']['token']

    def id(self):
        if 'login' not in self.cfg:
            raise RuntimeError('No token found in ' + self.fn +
                               ", please run 'ballsy login'")
        else:
            return self.cfg['login']['id']



from setuptools import find_packages, setup

try:
     import pypandoc
     description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError) as e:
     description = open('README.md').read()

dependencies = ['click', 'python-gnupg', 'github3.py', 'six']

setup(
    name='ballsy',
    version='0.1.0',
    url='https://github.com/satta/ballsy',
    license='BSD',
    author='Sascha Steinbiss',
    author_email='sascha@steinbiss.name',
    description='GitHub release signing tool',
    long_description=description,
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=dependencies,
    setup_requires=['pypandoc'],
    entry_points="""
        [console_scripts]
        ballsy=ballsy.cli:main
    """,
    classifiers=[
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ]
)
